@@ -5,17 +5,25 @@ sub _clean_eval { eval $_[0] }
 use strict;
 use warnings;
 
+our $VERSION = '2.006_007';
+$VERSION =~ tr/_//d;
+
 use Sub::Defer qw(defer_sub);
 use Scalar::Util qw(weaken);
-use Exporter qw(import);
+use Exporter ();
+BEGIN { *import = \&Exporter::import }
 use Carp qw(croak);
 BEGIN { our @CARP_NOT = qw(Sub::Defer) }
 use B ();
 BEGIN {
-  *_HAVE_IS_UTF8 = defined &utf8::is_utf8 ? sub(){1} : sub(){0};
-  *_HAVE_PERLSTRING = defined &B::perlstring ? sub(){1} : sub(){0};
-  *_BAD_BACKSLASH_ESCAPE = _HAVE_PERLSTRING() && "$]" == 5.010_000 ? sub(){1} : sub(){0};
-  *_HAVE_HEX_FLOAT = !$ENV{SUB_QUOTE_NO_HEX_FLOAT} && "$]" >= 5.022 ? sub(){1} : sub(){0};
+  my $TRUE  = sub(){!!1};
+  my $FALSE = sub(){!!0};
+  *_HAVE_IS_UTF8          = defined &utf8::is_utf8 ? $TRUE : $FALSE;
+  *_HAVE_PERLSTRING       = defined &B::perlstring ? $TRUE : $FALSE;
+  *_CAN_TRACK_BOOLEANS    = defined &builtin::is_bool ? $TRUE : $FALSE;
+  *_CAN_TRACK_NUMBERS     = defined &builtin::created_as_number ? $TRUE : $FALSE;
+  *_BAD_BACKSLASH_ESCAPE  = _HAVE_PERLSTRING() && "$]" == 5.010_000 ? $TRUE : $FALSE;
+  *_HAVE_HEX_FLOAT        = !$ENV{SUB_QUOTE_NO_HEX_FLOAT} && "$]" >= 5.022 ? $TRUE : $FALSE;
 
   # This may not be perfect, as we can't tell the format purely from the size
   # but it should cover the common cases, and other formats are more likely to
@@ -37,9 +45,6 @@ BEGIN {
   *_FLOAT_PRECISION = sub(){$precision};
 }
 
-our $VERSION = '2.006_005';
-$VERSION =~ tr/_//d;
-
 our @EXPORT = qw(quote_sub unquote_sub quoted_from_sub qsub);
 our @EXPORT_OK = qw(quotify capture_unroll inlinify sanitize_identifier);
 
@@ -63,11 +68,23 @@ if (_BAD_BACKSLASH_ESCAPE) {
 sub quotify {
   my $value = $_[0];
   no warnings 'numeric';
+  BEGIN {
+    warnings->unimport(qw(experimental::builtin))
+      if _CAN_TRACK_BOOLEANS || _CAN_TRACK_NUMBERS;
+  }
   ! defined $value     ? 'undef()'
+  : _CAN_TRACK_BOOLEANS && builtin::is_bool($value) ? (
+    $value ? '(!!1)' : '(!!0)'
+  )
   # numeric detection
-  : (!(_HAVE_IS_UTF8 && utf8::is_utf8($value))
-    && length( (my $dummy = '') & $value )
-    && 0 + $value eq $value
+  : (
+    _CAN_TRACK_NUMBERS
+      ? builtin::created_as_number($value)
+      : (
+        !(_HAVE_IS_UTF8 && utf8::is_utf8($value))
+        && length( (my $dummy = '') & $value )
+        && 0 + $value eq $value
+      )
   ) ? (
     $value != $value ? (
       $value eq (9**9**9*0)
@@ -90,12 +107,14 @@ sub quotify {
       FACTOR: for my $ex (0 .. abs($max_factor)) {
         my $num = $value / 2**($ex_sign * $ex);
         for my $precision (_FLOAT_PRECISION .. _FLOAT_PRECISION+2) {
-          my $formatted = sprintf '%.'.$precision.'g', $num;
+          my $formatted = sprintf '%0.'.$precision.'g', $num;
+          $float = $formatted
+            if $ex == 0;
           if ($formatted == $num) {
-            $float = $formatted;
             if ($ex) {
               $float
-                .= ($ex_sign == 1 ? '*' : '/')
+                = $formatted
+                . ($ex_sign == 1 ? '*' : '/')
                 . (
                   $ex > _NVMANTBITS
                     ? "2**$ex"
@@ -113,7 +132,7 @@ sub quotify {
       "$float";
     }
   )
-  : !length($value) && length( (my $dummy2 = '') & $value ) ? '(!1)' # false
+  : !_CAN_TRACK_BOOLEANS && !length($value) && length( (my $dummy2 = '') & $value ) ? '(!!0)' # false
   : _BAD_BACKSLASH_ESCAPE && _HAVE_IS_UTF8 && utf8::is_utf8($value) ? do {
     $value =~ s/(["\$\@\\[:cntrl:]]|[^\x00-\x7f])/
       $escape{$1} || sprintf('\x{%x}', ord($1))
@@ -266,6 +285,7 @@ sub _context {
       ."  \%^H = (\n"
       . join('', map
       "    ".quotify($_)." => ".quotify($hintshash->{$_}).",\n",
+        grep !(ref $hintshash->{$_} && $hintshash->{$_} =~ /\A(?:\w+(?:::\w+)*=)?[A-Z]+\(0x[[0-9a-fA-F]+\)\z/),
         keys %$hintshash)
       ."  );\n"
       ."}\n"
@@ -532,9 +552,12 @@ arguments.
 
  my $quoted_value = quotify $value;
 
-Quotes a single (non-reference) scalar value for use in a code string.  Numbers
-aren't treated specially and will be quoted as strings, but undef will quoted as
-C<undef()>.
+Quotes a single (non-reference) scalar value for use in a code string.  The
+result should reproduce the original value, including strings, undef, integers,
+and floating point numbers.  The resulting floating point numbers (including
+infinites and not a number) should be precisely equal to the original, if
+possible.  The exact format of the resulting number should not be relied on, as
+it may include hex floats or math expressions.
 
 =head2 capture_unroll
 
